
/*
 *	Masm
 *	Assembler for Merlin
 *
 *	Copyright 1995 VM Labs
 *
 *	Version 0.1
 *	
 *	Date started:	May 1st, 1995
 *	Current date:	May 28th, 1995
 */

/*	----------------------------------------------------------
 *	INCLUDES
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "errcode.h"


/*	----------------------------------------------------------
 * 	DEFINES
 */

#define DEBUG		1		/* let's debug		*/
#define LBUF_SIZE	1024		/* Line buffer size 	*/

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

/*	----------------------------------------------------------
 *	PROTOTYPES
 */

void	main(int, char **);
int	masm(void);
int	parse_instr(char *, long);
char	*parse_cmd(char *, long);
int	store_label(char *, long);
void	prt_error(long, int);
char	*mstrtok(char *, char *);

/*	----------------------------------------------------------
 * 	EXTERNALS
 */

extern char	*optarg;	/* std. option argument handling */
extern int 	optind;
extern int	opterr;
 
/*	----------------------------------------------------------
 * 	GLOBALS
 */

char		optstr[] = "o:";

char 		*prgName;	/* Name of program 		*/
char		*outName;	/* Output file name ptr 	*/
char		nameBuf[64];	/* Output file name buffer 	*/

FILE		*ifp;		/* Input file ptr		*/
FILE		*ofp;		/* Output file ptr		*/

char		lineBuf[LBUF_SIZE]; /* Line Buffer		*/

/*
	to do (6/13/95): equ table
 */

#define NAME_LEN	32
#define VALUE_LEN	128
#define N_EQU		512

typedef struct _t_equ T_EQU;

struct _t_equ {
	char	name[NAME_LEN];
	char	value[VALUE_LEN];
};

T_EQU equ_table[N_EQU];
int last_equ;


/*	---------------------------------------
 * 	Label/symbol management data structures
 */
#define LABEL_LEN	32

typedef struct _t_label T_LABEL;

struct _t_label {
	char	label[LABEL_LEN+1];
	long	addr;
	long	lineNo;
	T_LABEL	*next;
};

T_LABEL	*lastLabel;
T_LABEL	labelHead;
long	noLabels;

/*	-----------------
 *	instruction table
 */
#define N_INSTRS	128

#if 0
typedef struct _t_instr T_INSTR;

struct _t_instr {
	char	*name;		/* name of instruction */
	int	buildin;	/* flag for build in instr */
	int	nl_parameters;	/* nr of parameters to the left */
	int	nr_parameters;	/* nr of parameters to the right */
	int	addressmode;	/* address modi */
};

T_INSTR	instrs[N_INSTRS] = {
/*	  Name		BuildIn L.Par.	R.Par.	Address Mode	*/
	{ "include",	TRUE,	0,	1,	0 },
	{ "equ",	TRUE,	1,	1,	0 },
	{ NULL,		0,	0,	0,	0 },

};

#endif



/* ----------	
 * File stack
 */
#define N_FILES	128

typedef	struct _t_file T_FILE;

struct _t_file {
	char	name[256];	/* file name */
	FILE	*fp;		/* file ptr */
	int	line_no;		/* current line number in that file */
	int	in_comment;
	int	in_bracket;
};

T_FILE	files[N_FILES];
int	curr_file;		/* index to current file */


/*
 *	------------
 *	Lines
 */
#define BREAK_CHARS	" \t"

typedef struct _t_line T_LINE;

struct _t_line {
	T_LINE	*next;			/* next line */
	int	file_ix;		/* which file */
	int	abs_line_no;		/* absolute line number */	
	int	line_no;		/* line number in that file */
	int	pass_1;			/* flag set if pass 1 took care of it */
	int	n_token;		/* nr of tokens in this line */
	char	*token[64];
};

T_LINE	first_line;
T_LINE	*curr_line;
T_LINE	*last_line;

/*
 *	------------
 *	Etc.
 */
int	abs_line_no;
int	in_comment;		/* flag which shows us we are inside a comment */
int	conditional_code;	/* flag which shows we are inside of conditional code */
int	conditional_level;
long	curr_addr;		/* current output address */

/*
 *  Error messages
 */

char *errStrs[] = {
	"Colon inside instruction field.",
	"Double colon inside instruction field",
};


/*	----------------------------------------------------------
 *	FUNCTIONS
 */

	void
main(argc, argv)
	int argc;
	char **argv;
{
	int i;
	int code;	/* exit code */
	char c, *s;

	prgName = argv[0];
	outName = NULL;
	optind  = 1;
	code    = 0;
	
	/*
	 * Parse command line options
	 */
	while ((c = getopt(argc, argv, optstr)) != -1) {
		switch (c) {
		case 'o':
			outName = optarg;
			break;
		}
	}

	printf("masm version 0.1\n");

	/*
	 * Open input files & assemble them
	 */
	optind = (optind < 1) ? 1 : optind;
	for (i = optind; i < argc; i++) {	
printf("file=%s\n", argv[i]);
		if ((ifp = fopen(argv[i], "r")) == NULL) {
			perror(argv[i]);
			code = -1;
			continue;
		}
		/*
		 * Output name
		 */
		if (outName == NULL) {
			strcpy(nameBuf, argv[i]);
			if ((s = strrchr(nameBuf, '.')) != NULL) {

				strcat(++s, "o");
			} else {
				s = nameBuf;
				strcat(s, ".o");
			}
			outName = nameBuf;
		}
		/*
		 * Open output file name
		 */
#ifndef DEBUG
		if ((ofp = fopen(outName, "wb")) == NULL) {
#else
		if ((ofp = fopen(outName, "w")) == NULL) {
#endif
			perror(outName);
			code = -1;
			continue;
		}
		/*
		 * Assemble file...
		 */
		curr_file = 0;
		abs_line_no = 1;
		last_equ = 0;		
 		files[curr_file].line_no = 1;
 		files[curr_file].fp = ifp;
		strcpy(files[curr_file].name, argv[i]);
		curr_line = &first_line;

		if ((code = masm()) < 0)
			break;

		fclose(ofp);
	}
	
	exit(code);
}

/*
 *	Masm
 *	Main function to assemble a merlin assembly file
 */
	int
masm()

{
	int r1, r2, r3;

	r1 = pass_1();
	r2 = pass_2();
	r3 = pass_3();


}

/*
 *	1st pass, look for equates, globals etc.
 */

	int
pass_1()
{
	int r = 1;

	do {
		while (fgets(lineBuf, LBUF_SIZE, files[curr_file].fp) != NULL) {
			if (r == 1) {
				if ((curr_line->next = malloc(sizeof(T_LINE))) == NULL) {
					fprintf(stderr, "Out of memory.\n");
					return -1;
				}		
				curr_line = curr_line->next;
				curr_line->next = NULL;
			}
 			curr_line->file_ix = curr_file;
	 		curr_line->line_no = files[curr_file].line_no++;
 			curr_line->abs_line_no = abs_line_no++;
			curr_line->n_token = 0;
			curr_line->pass_1 = FALSE;

			r = scan_line(lineBuf);

			if (r == 1)
				if (pass_1_parser() == -1) {
					/*
					 * we are inside of conditional code
					 * so skip line
					 */
					r = 0;
				}
		}
		fclose(files[curr_file].fp);
		curr_file--;
	} while (curr_file >= 0);	
	return r;
}

	int
scan_line(linep)
	char	*linep;
{
	register char *s, *t;
	char *token;
	char x;
	static char line_copy[LBUF_SIZE];

	line_copy[0] = '\0';
	/*
	 * first, get rid of leading spaces
	 */
	for (s = linep; *s && isspace(*s); s++);	
	
	/*
	 * are we still in a comment section? 
	 */
	if (in_comment) {
		/*	look for end of comment
		 */
		for (; *s; s++) {
			if ((*s == '*') && (*(s + 1) == '/')) {
				in_comment = FALSE;
				s+=2;
				break; 
			}
		}
		if (!*s)
			return 0;	/* RETURN CODE HERE ??? */
	}

	/*
	 * next get rid of comments
	 */
	for (t = line_copy; *s; s++) {
		if (!in_comment && (*s == ';'))
			break;
		if ((*s == '/') * (*(s + 1) == '*')) {
			s+=1;
			in_comment = TRUE;
			continue;
		}
		if (!in_comment && (*s != '\n'))
			*t++ = *s;
		else if ((*s == '*') && (*(s + 1) == '/')) {
			s+=1;
			in_comment = FALSE;
		}
	}
	*t = '\0';

	if (in_comment)
		return 0;	/* RETURN CODE HERE ? */


	token = mstrtok(line_copy, BREAK_CHARS);
	while (token != NULL) {
		if ((t = malloc(strlen(token)+1)) == NULL) {
			fprintf(stderr, "Out of memory.\n");	
			return -1;
		}
		strcpy(t, token);
		curr_line->token[curr_line->n_token++] = t;

		token = mstrtok(NULL, BREAK_CHARS);
	}
	
	
	return 1;
}

/*
 *	Check for pass 1 commands
 */
#define P1_INCLUDE 	0
#define P1_IF	   	1
#define P1_ELSE		2
#define P1_ELIF		3
#define P1_ENDIF	4
#define P1_EQU		5

	int
pass_1_parser()
{
	int	cmd, spot;


	cmd = pass_1_commands(&spot);

	if (conditional_code) {
		if ((cmd != P1_ELSE) && (cmd != P1_ENDIF) && (cmd != P1_ELIF))
			return -1;
	}

	switch (cmd) {
	case P1_INCLUDE:
		p1_cmd_include(spot);
		break;
	case P1_IF:
		p1_cmd_if(spot);
		break;
	case P1_ELSE:
		p1_cmd_else(spot);
		break;
	case P1_ENDIF:
		p1_cmd_endif(spot);
		break;
	case P1_EQU:
		p1_cmd_equ(spot);
		break;
	}
	return 0;
}

/*
 *	check for pass 1 commands
 */

char	*p1_cmds[] = {
	"include",
	"if",
	"else",
	"elif",
	"endif",
	"equ",
	"extern",
	NULL
};

	int
pass_1_commands(spot)
	int	*spot;
{	
	register int i, k;
	register char **s, *t;

	for (s = p1_cmds, k = 0; *s; s++, k++) {
		for (i = 0; i < curr_line->n_token; i++) {
			t = (curr_line->token[i][0] == '.') ?
				&curr_line->token[i][1] :
				&curr_line->token[i][0];
			if (strcmp(t, *s) == 0) {
				curr_line->pass_1 = TRUE;
				*spot = i;
				return k;
			}
		}
	}

	return -1;

}

/*
 *	include statement
 */

p1_cmd_include(spot)
	int spot;
{
	FILE *fp;
	register char *s, *t;
	char	fname[256];

	if (spot != 0) {
		fprintf(stderr, "nothing allowed in front of include\n");
		return -1;
	}
	if (curr_line->token[1] == NULL) {
		fprintf(stderr, "file name missing\n");
		return -1;
	}
	if (curr_line->token[1][0] != '\"') {
		fprintf(stderr, "\" missing\n");
		return -1;
	}
	for (t = fname, s = &curr_line->token[1][1]; *s && *s != '\"'; )
		*t++ = *s++;
	*t = '\0';
	if (*s != '\"') {
		fprintf(stderr, "\" missing\n");
		return -1;
	}
	
	if ((fp = fopen(fname, "r")) == NULL) {
		fprintf(stderr, "can\'t open include file %s\n", fname);
		return -1;
	}
	curr_file++;
	files[curr_file].fp = fp;
	strcpy(files[curr_file].name, fname);
 	files[curr_file].line_no = 1;

	return 0;
}

/*
 *	equ statement
 */

p1_cmd_equ(spot)
	int spot;
{
	FILE *fp;
	register char *s, *t;

	if (spot != 1) {
		fprintf(stderr, "nothing in front of equ. syntax error.\n");
		return -1;
	}
	if (curr_line->token[0] == NULL) {
		fprintf(stderr, "variable missing\n");
		return -1;
	}
	if (curr_line->token[2] == NULL) {
		fprintf(stderr, "expression missing\n");
		return -1;
	}	

	if (check_dup_equs(curr_line->token[1]) == TRUE) {
		fprintf(stderr, "duplicate equate\n");
		return -1;
	}	
	strncpy(equ_table[last_equ].name, curr_line->token[0], 32);
	strncpy(equ_table[last_equ].value, curr_line->token[2], 128);


#if	DEBUG
	printf("equ: %s = %s\n", equ_table[last_equ].name, equ_table[last_equ].value);
#endif

	last_equ++;

	return 0;
}

	int
check_dup_equs(name)
	char *name;
{
	register int i;

	for (i = 0; i < last_equ; i++)
		if (strncmp(name, equ_table[i].name, 32) == 0)
			return TRUE;
	return FALSE;
}

/*
 *	if statement
 */

p1_cmd_if(spot)
	int spot;
{
	FILE *fp;
	register char *s, *t;

	if (spot != 0) {
		fprintf(stderr, "nothing allowed in front of if\n");
		return -1;
	}
	if (curr_line->token[1] == NULL) {
		fprintf(stderr, "expression missing\n");
		return -1;
	}
	/*
	 *	test code for now, LATER we have to call expression builder
	 */

	if (curr_line->token[1][0] == '1')
		conditional_code = FALSE;	
	else if (curr_line->token[1][0] == '0')
		conditional_code = TRUE;
	conditional_level += 1;

}

/*
 *	else statement
 */

p1_cmd_else(spot)
	int spot;
{
	FILE *fp;
	register char *s, *t;

 	if (spot != 0) {
		fprintf(stderr, "nothing allowed in front of else\n");
		return -1;
	}
	if (conditional_level < 1) {
		fprintf(stderr, "else without an if\n");
		return -1;
	}

	if (curr_line->n_token > 1) {
		fprintf(stderr, "nothing allowd after else\n");
		return -1;
	}
	conditional_code = (!conditional_code);	

	return 0;
}

/*
 *	endif statement
 */

p1_cmd_endif(spot)
	int spot;
{
	FILE *fp;
	register char *s, *t;

	if (spot != 0) {
		fprintf(stderr, "nothing allowed in front of endif\n");
		return -1;
	}
	if (conditional_level < 1) {
		fprintf(stderr, "endif without an if\n");
		return -1;
	}

	if (curr_line->n_token > 1) {
		fprintf(stderr, "nothing allowd after endif\n");
		return -1;
	}

	conditional_code = FALSE;
	conditional_level--;	

	return 0;
}




pass_2()
{
	register T_LINE *lp;
	register int i;

	for (lp = first_line.next; lp != NULL; lp = lp->next) {
		for (i = 0; i < lp->n_token; i++) {
			printf("%s | ", lp->token[i]);
		}
		printf("\n");
	}
}

pass_3()
{

}


#if 0 	/*old stuff */
/*
 *	Masm
 *	Main function to assemble a merlin assembly file
 */
	int
parse_line(lineBuf)
	char *lineBuf;
{
	int error;
	long lineNo;
	int inComment;
	int inSpace;
	int inInstr;
	char *afterSpace;
	register char *s;


	curr_addr = 0x0L;
	lineNo    = 0L;
	inComment = FALSE;
	inInstr   = FALSE;

	noLabels  = 0L;
	lastLabel = &labelHead;


	afterSpace = lineBuf;
	inSpace    = FALSE;

	/*
	 * Go through line
	 */
	for (s = lineBuf; *s; s++) {
		/*
		 * First, do comments
		 */
		if (inComment) {
			if ((*s == '*') && (*(s+1) == '/')) {
				inComment = FALSE;
				s++;
			}
			continue;
		}	
		if (*s == ';')
			break;
		if ((*s == '/') && (*(s + 1) == '*')) {
			inComment = TRUE;
			s++;
			continue;
		}
		/*
		 * Spaces
		 */
		if (isspace(*s)) {
			s = parse_instr(afterSpace, lineNo);
			inSpace =TRUE;
			continue;
		} 
		if (inSpace) {
			afterSpace = s;
			inSpace = FALSE;
		}
		/*
		 * Labels
		 */
		if (*s == ':') {
			if (inInstr) {
				prt_error(lineNo, ERR_DCOLON_IN_INSTR);
				break;
			}
			*s = '\0';
			store_label(afterSpace, lineNo);
		}
			

	}
	lineNo++;


	return error;
}


/*
 * Parse a merlin instr.
 */

	int
parse_instr(instr, lineNo)
	char *instr;
	long lineNo;
{
	instr_ix = find_instr(instr);


	return 0;
}

	int
find_instr(instr)
	char *instr;
{
	for (i = 0; t != NULL; i++, t++)
		
}

	char *
parse_cmd(cmd, lineNo)
	char *cmd;
	long lineNo;
{
#ifdef DEBUG
	fprintf(ofp, "%ld: .%s\n", lineNo, cmd);
#endif

	return cmd;
}

/*
 *	Store label into table
 */

	int
store_label(label, lineNo)
	char *label;
	long lineNo;
{
#ifdef DEBUG
	fprintf(ofp, "%ld: %s\n", lineNo, label);
#endif
	if ((lastLabel->next = (T_LABEL*) malloc(sizeof(T_LABEL))) == NULL) {
		printf("Out of memory.\n");
		exit(-1);
	}
	lastLabel = lastLabel->next;
	strncpy(lastLabel->label, label, LABEL_LEN);
	lastLabel->addr   = curr_addr;
	lastLabel->lineNo = lineNo;

	noLabels += 1L;

	return 0;
}
#endif

/*
 * print error message
 */
	void
prt_error(lineNo, errno)
	long lineNo;
	int errno;
{
	fprintf(stderr, "%s:\t(line %d) %s\n", outName, lineNo, errStrs[errno]);
 
}


	char
*mstrtok(line, brkchrs)
	char *line;
	char *brkchrs;
{
	static char *s;
	static char x; 
	static int i, len;
	int in_string;
	char *s0, *t;
	int f;


	if (line != NULL) {
		s = line;
		len = strlen(s);
		i = 0;
	} else {
		*s++ = x;
		i++;
	}

	if (!*s || (i >= len))
		return NULL;

	in_string = FALSE;

	for (; *s; s++, i++) {
		for (f = 0, t = brkchrs; *t; t++) {
			if (*s == *t) {
				f = 1;
				break;
			}
		}
		if (!f)
			break;
	}
	for (s0 = s; *s; s++, i++) {
		if (in_string && (*s == '\"')) {
			if ((i == 0) || ((i > 0) && (*(s-1) != '\\'))) {
				in_string = FALSE;
				x = *++s;
				*s = '\0';
				break;
			}
		}
		if (in_string)
			continue;
		
		if (*s == '\"')
			in_string = TRUE;


		for (t = brkchrs; *t; t++) {
			if (*s == *t) {
				x = *s;
				*s = '\0';
				break;
			}
		}
		if (!*s)
			break;
	}		
	if (in_string)
		fprintf(stderr, "missing \"\n");
	
	return s0;
}
